import asyncio
from httpx import AsyncClient
import pytest
import json
from functools import reduce
from PLATER.services.util.graph_adapter import GraphInterface
import os

from PLATER.services.app import APP, get_graph_interface


class MockGraphInterface(GraphInterface):
    def __init__(self, *args, **kwargs):
        pass

    def get_schema(self):
        graph_schema_file_path = os.path.join(os.path.dirname(__file__), 'data', 'graph_schema.json')
        with open(graph_schema_file_path) as j_file:
            return json.load(j_file)

    async def get_mini_schema(self, source_id, target_id):
        #this function is only used by simple_spec endpoint
        # we could assert that its being called appropriately
        if source_id:
            assert source_id == 'SOME:CURIE'
        if target_id:
            assert target_id == 'SOME:CURIE'

        schema = self.get_schema()
        # flatten the schema, to mimic
        # MATCH (a)-[e]->(b) return labels(a) as source_label, type(e) as predicate, labels(b) as target_label
        flat_schema = []
        for source_type in schema:
            for target_type in schema[source_type]:
                for edge_type in schema[source_type][target_type]:
                    flat_schema.append({
                        'source_label': [source_type],
                        'predicate': edge_type,
                        'target_label': [target_type]})
        return flat_schema

    async def get_node(self, node_type, curie):
        node_list_file_path = os.path.join(os.path.dirname(__file__), 'data', 'node_list.json')
        with open(node_list_file_path) as j_file:
            return json.load(j_file)

    async def get_single_hops(self, source_type, target_type, curie):
        single_hop_triplets_file_path = os.path.join(os.path.dirname(__file__), 'data', 'single_hop_triplets.json')
        with open(single_hop_triplets_file_path) as j_file:
            return json.load(j_file)

    async def run_cypher(self, cypher, return_errors=False):

        return {
            'results': [],
            "errors": []
        }

    async def get_examples(self, source, target=None):
        single_hop_triplets_file_path = os.path.join(os.path.dirname(__file__), 'data', 'single_hop_triplets.json')
        with open(single_hop_triplets_file_path) as j_file:
            triplets = json.load(j_file)
        return reduce(lambda x, y: x + [y[0]], triplets, [])


def _graph_interface():
    return MockGraphInterface('host', 'port', ('neo4j', 'pass'))


@pytest.fixture()
def graph_interface():
    return _graph_interface()


APP.dependency_overrides[get_graph_interface] = _graph_interface


@pytest.mark.asyncio
async def test_node_response(graph_interface):
    async with AsyncClient(app=APP, base_url="http://test") as ac:
        response = await ac.get("/chemical_substance/curie")
    assert response.status_code == 200
    graph_response = await graph_interface.get_node('chemical_substance', 'curie')
    assert response.json() == graph_response


@pytest.mark.asyncio
async def test_one_hop_response(graph_interface):
    async with AsyncClient(app=APP, base_url="http://test") as ac:
        response = await ac.get("/chemical_substance/gene/CHEBI:11492")
    assert response.status_code == 200
    graph_response = await graph_interface.get_single_hops('chemical_substance', 'gene', 'CHEBI:11492')
    assert response.json() == graph_response


@pytest.mark.asyncio
async def test_cypher_response(graph_interface):
    query = 'MATCH (n) return n limit 1'
    async with AsyncClient(app=APP, base_url="http://test") as ac:
        response = await ac.post("/cypher", json={
            "query": query
        })
    assert response.status_code == 200
    graph_resp = await graph_interface.run_cypher(query)
    assert response.json() == graph_resp


@pytest.mark.asyncio
async def test_graph_schema_response(graph_interface):
    async with AsyncClient(app=APP, base_url="http://test") as ac:
        response = await ac.get("/graph/schema")
    assert response.status_code == 200
    assert response.json() == graph_interface.get_schema()


@pytest.mark.asyncio
async def test_simple_one_hop_spec_response(graph_interface):
    # with out parameters it should return all the questions based on that
    # send source parameter, target parameter
    async with AsyncClient(app=APP, base_url="http://test") as ac:
        response = await ac.get("/simple_spec")
    assert response.status_code == 200
    specs = response.json()
    schema = graph_interface.get_schema()
    source_types = set(schema.keys())
    target_types = set(reduce(lambda acc, source: acc + list(schema[source].keys()), schema, []))
    spec_len = 0
    for source in schema:
        for target in schema[source]:
            spec_len += len(schema[source][target])
    assert len(specs) == spec_len

    for item in specs:
        assert item['source_type'] in source_types
        assert item['target_type'] in target_types

    # test source param
    source_type = list(schema.keys())[0]
    async with AsyncClient(app=APP, base_url="http://test") as ac:
        response = await ac.get("/simple_spec?source=SOME:CURIE")
    # response = client.get(f'/simple_spec?source=SOME:CURIE')
    assert response.status_code == 200
    async with AsyncClient(app=APP, base_url="http://test") as ac:
        response = await ac.get("/simple_spec?source=SOME:CURIE")
    # response = client.get(f'/simple_spec?target=SOME:CURIE')
    assert response.status_code == 200



from PLATER.services.util.graph_adapter import GraphInterface
from PLATER.services.util.question import Question
from functools import reduce


class Overlay:
    def __init__(self, graph_interface: GraphInterface):
        self.graph_interface = graph_interface

    async def overlay_support_edges(self, reasoner_graph):
        """
        Grabs a set of answers and queries for connection among set of nodes
        :param reasoner_graph:
        :return:
        """
        final_response = {}
        edges_to_add = dict()
        overlayed_answers = list()
        chunk_size = 1000
        chunked_answers = [reasoner_graph[Question.ANSWERS_KEY][start: start + chunk_size]
                           for start in range(0, len(reasoner_graph[Question.ANSWERS_KEY]), chunk_size)]
        for answer in chunked_answers:
            # 3. filter out kg ids
            all_kg_nodes = []
            for ans in answer:
                ans_nodes = ans[Question.NODE_BINDINGS_KEY]
                for qid in ans_nodes:
                    nodes = ans_nodes[qid]
                    for n in nodes:
                        all_kg_nodes.append(n['id'])

            # all_kg_nodes = set(
            #     reduce(lambda a, b: a + b,
            #            map(lambda node_binding: node_binding[Question.KG_ID_KEY]
            #            if isinstance(node_binding[Question.KG_ID_KEY], list)
            #            # 3. convert every thing to array and reduce
            #            else [node_binding[Question.KG_ID_KEY]],
            #                # 2. merge them into single array
            #                reduce(lambda a, b: a + b,
            #                       # 1. get node bindings from all answers
            #                       map(lambda ans: ans[Question.NODE_BINDINGS_KEY], answer), [])), []))
            # fun part summon APOC
            if self.graph_interface.supports_apoc():
                all_kg_nodes = list(all_kg_nodes)
                apoc_result = (await self.graph_interface.run_apoc_cover(all_kg_nodes))[0]['result']
                apoc_result = self.structure_for_easy_lookup(apoc_result)
                # now go back to the answers and add the edges
                for ans in answer:
                    support_id_suffix = 0
                    node_bindings = ans[Question.NODE_BINDINGS_KEY]
                    ans_all_node_ids = set()
                    for qid in node_bindings:
                        nodes = node_bindings[qid]
                        for n in nodes:
                            ans_all_node_ids.add(n['id'])
                    for node_id in ans_all_node_ids:
                        other_nodes = ans_all_node_ids.difference(set([node_id]))
                        # lookup current node in apoc_result
                        current_node_relations = apoc_result.get(node_id, {})
                        for other_node_id in other_nodes:
                            # lookup for relations in apoc_result graph
                            support_edges = current_node_relations.get(other_node_id, [])
                            for support_edge in support_edges:
                                q_graph_id = f's_{support_id_suffix}'
                                support_id_suffix += 1
                                k_graph_id = support_edge['id']
                                del support_edge['id']
                                ans['edge_bindings'][q_graph_id] = [{"id": k_graph_id}]
                                if k_graph_id not in edges_to_add:
                                    edges_to_add[k_graph_id] = support_edge
                    overlayed_answers.append(ans)
                    # @TODO raise exception if apoc is not supported

        final_response[Question.QUERY_GRAPH_KEY] = reasoner_graph[Question.QUERY_GRAPH_KEY]
        final_response[Question.ANSWERS_KEY] = overlayed_answers
        final_response[Question.KNOWLEDGE_GRAPH_KEY] = reasoner_graph[Question.KNOWLEDGE_GRAPH_KEY]
        final_response[Question.KNOWLEDGE_GRAPH_KEY][Question.EDGES_LIST_KEY].update(edges_to_add)
        return final_response

    def structure_for_easy_lookup(self, result_set):
        """
        Converts apoc result into a mini graph
        :param result_set:
        :return:
        """
        result = {}
        for r in result_set:
            source_id = r['subject']
            target_id = r['object']
            edge = {}
            edge['subject'] = source_id
            edge['object'] = target_id
            edge['predicate'] = r['predicate']
            edge['id'] = r['edge']['id']
            edge['attributes'] = [{
                "type": "WIKIDATA:Q80585",
                "value": r['edge']
            }]
            m = result.get(source_id, {})
            n = m.get(target_id, list())
            n.append(edge)
            m[target_id] = n
            result[source_id] = m
        return result




###############
#  Configuration loader
#  Borrowed from : https://github.com/NCATS-Gamma/robokop-interfaces/blob/master/greent/config.py
#
##################
import os
import yaml
import traceback
import re


class Config(dict):

    @staticmethod
    def get_resource_path(resource_name):
        """ Given a string resolve it to a module relative file path unless it is already an absolute path. """
        resource_path = resource_name
        if not resource_path.startswith(os.sep):
            resource_path = os.path.join(os.path.dirname(__file__), resource_path)
        return resource_path

    def __init__(self, config, prefix=''):
        '''
        if not config.startswith (os.sep):
            config = os.path.join (os.path.dirname (__file__), config)
        '''
        if isinstance(config, str):
            config_path = Config.get_resource_path(config)
            with open(config_path, 'r') as f:
                self.conf = yaml.safe_load (f)
        elif isinstance(config, dict):
            self.conf = config
        else:
            raise ValueError
        self.prefix = prefix

    def get_service(self, service):
        result = {}
        try:
            result = self['translator']['services'][service]
        except:
            traceback.print_exc()
        return result

    def __setitem__(self, key, val):
        raise TypeError("Setting configuration is not allowed.")

    def __str__(self):
        return "Config with keys: "+', '.join(list(self.conf.keys()))

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __getitem__(self, key):
        """
        Use this accessor instead of getting conf directly in order to permit overloading with environment variables.
        Imagine you have a config file of the form

          person:
            address:
              street: Main

        This will be overridden by an environment variable by the name of PERSON_ADDRESS_STREET,
        e.g. export PERSON_ADDRESS_STREET=Gregson
        """
        key_var = re.sub("[\\W]", '', key)

        name = self.prefix+'_'+key_var if self.prefix else key_var
        try:
            env_name = name.upper()
            return os.environ[env_name]
        except KeyError:
            value = self.conf[key]
            if isinstance(value, dict):
                return Config(value, prefix=name)
            else:
                return value


config = Config('plater.conf')



"""Tools for compiling QGraph into Cypher query."""
import re
from reasoner_converter.downgrading import downgrade_BiolinkEntity, downgrade_BiolinkPredicate
from reasoner_converter.upgrading import upgrade_BiolinkEntity, upgrade_BiolinkRelation


def cypher_prop_string(value):
    """Convert property value to cypher string representation."""
    if isinstance(value, bool):
        return str(value).lower()
    elif isinstance(value, str):
        return f"'{value}'"
    else:
        raise ValueError(f'Unsupported property type: {type(value).__name__}.')


class NodeReference():
    """Node reference object."""

    def __init__(self, node, node_id ,anonymous=False):
        """Create a node reference."""
        node = dict(node)
        name = f'{node_id}' if not anonymous else ''
        labels = node.pop('category', 'named_thing')
        if not isinstance(labels, list):
            labels = [labels]
        props = {}
        curie_filters = []
        curie = node.pop("id", None)
        self.has_curie = False
        if curie is not None:
            if isinstance(curie, str) or (isinstance(curie, list) and len(curie) == 1):
                props['id'] = curie if isinstance(curie, str) else curie[0]
            elif isinstance(curie, list):
                for ci in curie:
                    # generate curie-matching condition
                    curie_filters.append(f"{name}.id = '{ci}'")
                # union curie-matching filters together
            else:
                raise TypeError("Curie should be a string or list of strings.")
            self.has_curie = True
        label_filters = []
        if labels:
            biolink_regex = "^biolink:[A-Z][a-zA-Z]*$"
            for label in labels:
                label_filters.append(f"'{label}' in {name}.category")
                is_biolinkified = re.match(biolink_regex, label)
                if is_biolinkified:
                    other_label =  downgrade_BiolinkEntity(label)
                else:
                    other_label = upgrade_BiolinkEntity(label)
                label_filters.append(f"'{other_label}' in {name}.category")
        self._filters = ''
        if len(curie_filters):
            filters = '( ' + ' OR '.join(curie_filters) + ')'
            self._filters = filters
        node.pop('name', None)
        node.pop('is_set', False)
        props.update(node)

        self.name = name
        self.labels = labels
        self.prop_string = ' {' + ', '.join([f"`{key}`: {cypher_prop_string(props[key])}" for key in props]) + '}'
        if curie:
            # redis graph doesnt support USING INDEX version 2.4.2
            self._extras = '' #f' USING INDEX {name}:`{labels[0]}`(id)'
        else:
            self._extras = ''
        self._num = 0

    def __str__(self):
        """Return the cypher node reference."""
        self._num += 1
        if self._num == 1:
            label = f':`{self.labels[0]}`' if self.labels else ''
            return f'{self.name}' + label + f'{self.prop_string}' # + ''.join(f':`{label}`' for label in self.labels)
        return self.name

    @property
    def filters(self):
        """Return filters for the cypher node reference.
        To be used in a WHERE clause following the MATCH clause.
        """
        return self._filters

    @property
    def extras(self):
        """Return extras for the cypher node reference.
        To be appended to the MATCH clause.
        """
        if self._num >= 1:
            return self._extras
        else:
            return ''


class EdgeReference:
    """Edge reference object."""

    def __init__(self, edge, edge_id, anonymous=False):
        """Create an edge reference."""

        name = f'{edge_id}' if not anonymous else ''
        label = edge['predicate'] if 'predicate' in edge else None

        if 'predicate' in edge and edge['predicate'] is not None:
            if isinstance(edge['predicate'], str):
                label = edge['predicate']
                filters = ''
            elif isinstance(edge['predicate'], list):
                filters = []
                biolink_predicate_regex = "^biolink:[a-z][a-z_]*$"
                for predicate in edge['predicate']:
                    is_biolink_predicate = re.match(biolink_predicate_regex, predicate)
                    filters.append(f'type({name}) = "{predicate}"')
                    if is_biolink_predicate:
                        other_predicate = downgrade_BiolinkPredicate(predicate)
                    else:
                        other_predicate = upgrade_BiolinkRelation(predicate)
                    filters.append(f'type({name}) = "{other_predicate}" ')
                filters = ' OR '.join(filters)
                label = None
        else:
            label = None
            filters = ''

        self.name = name
        self.label = label
        self._num = 0
        self._filters = filters
        has_type = 'predicate' in edge and edge['predicate']
        self.directed = edge.get('directed', has_type)
        self.reversed = False

    def __str__(self):
        """Return the cypher edge reference."""
        self._num += 1
        if self._num == 1:
            label = f':`{self.label}`' if self.label else ''
            innards = f'{self.name}{label}'
        else:
            innards = self.name
        if self.directed:
            return f'-[{innards}]->' if not self.reversed else f'<-[{innards}]-'
        else:
            return f'-[{innards}]-'

    @property
    def filters(self):
        """Return filters for the cypher node reference.
        To be used in a WHERE clause following the MATCH clause.
        """
        return self._filters


def cypher_query_fragment_match(qgraph, max_connectivity=-1):
    """Generate a Cypher query fragment to match the nodes and edges that correspond to a question.
    This is used internally for cypher_query_answer_map and cypher_query_knowledge_graph
    Returns the query fragment as a string.
    """
    nodes, edges = qgraph['nodes'], qgraph['edges']

    # generate internal node and edge variable names
    node_references = {n: NodeReference(node=nodes[n], node_id=n) for n in nodes}
    edge_references = {e: EdgeReference(edge=edges[e], edge_id=e) for e in edges}

    match_strings = []

    # match orphaned nodes
    def flatten(l):
        return [e for sl in l for e in sl]
    all_nodes = set(nodes.keys())
    all_referenced_nodes = set(flatten([[edges[e]['subject'], edges[e]['object']] for e in edges]))
    orphaned_nodes = all_nodes - all_referenced_nodes
    nodes_with_id = [n_id for n_id, node in nodes.items() if node.get('id')]
    for n in orphaned_nodes:
        match_strings.append(f"MATCH ({node_references[n]})")
        match_strings[-1] += node_references[n].extras
        if node_references[n].filters:
            match_strings.append("WHERE " + node_references[n].filters)
    nodes_so_far = []
    for n in nodes_with_id:
        match_strings.append(f"MATCH ({node_references[n]})")
        if node_references[n].filters:
            match_strings.append("WHERE " + node_references[n].filters)
        nodes_so_far.append(n)
        match_strings.append(f"WITH {', '.join(nodes_so_far)}")

    # match edges
    for edge_id, eref in edge_references.items():
        e = edges[edge_id]
        source_node = node_references[e['subject']]
        target_node = node_references[e['object']]
        if target_node.has_curie and not source_node.has_curie:
            eref.reversed = True
            match_strings.append(f"MATCH ({target_node}){eref}({source_node})")
        else:
            match_strings.append(f"MATCH ({source_node}){eref}({target_node})")
        match_strings[-1] += source_node.extras + target_node.extras
        filters = [f'({c})' for c in [source_node.filters, target_node.filters, eref.filters] if c]
        if max_connectivity > -1:
            filters.append(f"((indegree({target_node}) + outdegree({target_node}) ) < {max_connectivity})")
        if filters:
            match_strings.append("\nWHERE " + "\nAND ".join(filters))

    match_string = ' '.join(match_strings)

    return match_string


def cypher_query_answer_map(qgraph, **kwargs):
    """Generate a Cypher query to extract the answer maps for a question.
    Returns the query as a string.
    """
    clauses = []

    match_string = cypher_query_fragment_match(qgraph, max_connectivity=kwargs.pop('max_connectivity', -1))
    if match_string:
        clauses.append(match_string)

    nodes, edges = qgraph['nodes'], qgraph['edges']

    # generate internal node and edge variable names
    node_names = set(nodes.keys())
    node_names_sets = set(f"{n}" for n in filter(lambda n: nodes[n].get('is_set'), nodes))
    edge_names = set(edges.keys())

    # deal with sets
    node_id_accessor = [f"collect({n}) AS {n}" if n in node_names_sets
                        else f"{n} AS {n}" for n in node_names]
    edge_id_accessor = [f"collect({e}) AS {e}" for e in edge_names]
    if node_id_accessor or edge_id_accessor:
        with_string = f"WITH {', '.join(node_id_accessor+edge_id_accessor)}"
        clauses.append(with_string)

    returns = list(node_names) + \
              list(edge_names) + \
              [f'[node in {x} | labels(node)] AS type__{x}' for x in node_names_sets] + \
              [f'labels({x}) AS type__{x}' for x in node_names - node_names_sets] + \
              [f'[edge in {x} | type(edge)] AS type__{x}' for x in edge_names] + \
              [f'[edge in {x} | [startNode(edge).id, endNode(edge).id]] AS id_pairs__{x}' for x in edge_names]

    answer_return_string = f"RETURN " + ','.join(returns)

    clauses.append(answer_return_string)

    # return answer maps matching query
    query_string = '\n'.join(clauses)
    if 'skip' in kwargs:
        query_string += f' SKIP {kwargs["skip"]}'
    if 'limit' in kwargs:
        query_string += f' LIMIT {kwargs["limit"]}'
    return query_string


import redis
from PLATER.services.config import config
from PLATER.services.util.logutil import LoggingUtil
from PLATER.services.util.drivers.redis_trapi_cypher_compiler import cypher_query_answer_map
from redisgraph import Graph, Node, Edge
from reasoner_converter.upgrading import upgrade_BiolinkRelation, upgrade_BiolinkEntity

logger = LoggingUtil.init_logging(__name__,
                                  config.get('logging_level'),
                                  config.get('logging_format')
                                  )


class RedisDriver:
    def __init__(self, host, port=6379, password=None, graph_db_name='test'):
        self.redis_url = f'redis://:{password}@{host}:{port}' if password else f'redis://{host}:{port}'
        self.redis_client = None
        self.sync_redis_client = redis.Redis(host=host,
                                             port=port,
                                             password=password,
                                             encoding='utf-8',
                                             decode_responses=True)
        self.graph_name = graph_db_name
        self.redis_graph = Graph(self.graph_name, self.sync_redis_client)
        self.ping_redis()

    def ping_redis(self):
        logger.info('[x] Pinging redis')
        response = self.sync_redis_client.execute_command('ping')
        logger.info(f'[x] Got response...{response}')

    @staticmethod
    def format_cypher_result(redis_results):
        return {
            'results': [{
                'columns': redis_results[0],
                'data': [{'row': x, 'meta': []} for x in redis_results[1]]
            }],
            'errors': []
        }

    @staticmethod
    def decode_if_byte(value):
        try:
            return value.decode('utf-8')
        except:
            return value

    async def run(self, query, **kwargs):
        query_timeout = kwargs.get('query_timeout', None)
        results = self.redis_graph.query(query, read_only=True, timeout= query_timeout)
        headers = list(map(lambda x: RedisDriver.decode_if_byte(x[1]), results.header))
        response = []
        for row in results.result_set:
            new_row = []
            for value in row:
                if isinstance(value, list):
                    parsed_value = []
                    for v in value:
                        if isinstance(v, Node) or isinstance(v, Edge):
                            parsed_value.append(v.properties)
                        else:
                            parsed_value.append(v)
                    new_row.append(parsed_value)
                elif isinstance(value, Node) or isinstance(value, Edge):
                    new_row.append(value.properties)
                else:
                    new_row.append(value)
            response.append(new_row)
        return self.format_cypher_result((headers, response))

    def run_sync(self, cypher_query):
        results = self.sync_redis_client.execute_command('GRAPH.RO_QUERY', self.graph_name, cypher_query)
        return RedisDriver.format_cypher_result(results)

    @staticmethod
    def convert_to_dict(response: dict) -> list:
        """
        Converts a neo4j result to a structured result.
        :param response: neo4j http raw result.
        :type response: dict
        :return: reformatted dict
        :rtype: dict
        """
        results = response.get('results')
        array = []
        if results:
            for result in results:
                cols = result.get('columns')
                if cols:
                    data_items = result.get('data')
                    for item in data_items:
                        new_row = {}
                        row = item.get('row')
                        for col_name, col_value in zip(cols, row):
                            new_row[col_name] = col_value
                        array.append(new_row)
        return array

    def transplile_TRAPI_cypher(self, trapi_question, options={}):
        return cypher_query_answer_map(trapi_question, **options)

    async def answer_TRAPI_question(self, trapi_question, options={}, timeout=None):
        cypher = self.transplile_TRAPI_cypher(trapi_question, options)
        logger.info("RUNNING TRAPI QUERY: ")
        logger.info(cypher)
        results = await self.run(cypher, query_timeout=timeout)
        results_dict = self.convert_to_dict(results)
        response = self.create_TRAPI_kg_response(trapi_question, results_dict)
        return response

    def create_TRAPI_kg_response(self, query_graph , results_dict):
        node_qg_ids = list(query_graph['nodes'].keys())
        edge_qg_ids = list(query_graph['edges'].keys())
        answer_bindings = []
        nodes_all = {}
        edges_all = {}
        collected_nodes = set()
        collected_edges = set()

        for row in results_dict:
            current_answer_bindings = {
                'node_bindings': {},
                'edge_bindings': {}
            }
            bound_nodes = {}
            for qg_id in node_qg_ids:
                # Convert nodes and node types to list
                nodes = row[qg_id] if isinstance(row[qg_id], list) else [row[qg_id]]
                node_types = row[f'type__{qg_id}'] if isinstance(row[qg_id], list) else [row[f'type__{qg_id}']]
                current_node_binding = {qg_id: []}
                for node, node_type in zip(nodes, node_types):
                    node_id = node.pop('id')
                    assert node_id, 'Error, did not find ID from Node in db'
                    current_node_binding[qg_id].append({'id': node_id})
                    bound_nodes[qg_id] = bound_nodes.get(qg_id, [])
                    bound_nodes[qg_id].append(node_id)
                    if node_id not in collected_nodes:
                        new_node = {}
                        collected_nodes.add(node_id)
                        new_node['category'] = [upgrade_BiolinkEntity(x) for x in node['category']]
                        new_node['name'] = node.get('name', '')
                        new_node['attributes'] = []
                        for key, value in node.items():
                            if key in new_node:
                                continue
                            new_node['attributes'].append({
                                'type': 'NA',
                                'value': value,
                                'name': key
                            })
                        nodes_all[node_id] = new_node
                current_answer_bindings['node_bindings'].update(current_node_binding)
            for qg_id in edge_qg_ids:
                edges = row[qg_id] if isinstance(row[qg_id], list) else [row[qg_id]]
                edge_types = row[f'type__{qg_id}'] if isinstance(row[qg_id], list) else [row[f'type__{qg_id}']]
                id_pairs = row[f'id_pairs__{qg_id}']
                index = 0
                current_edge_binding = {qg_id: []}
                for edge, edge_type, id_pair in zip(edges, edge_types, id_pairs):
                    edge_id = edge.pop('id')
                    current_edge_binding[qg_id].append({'id': edge_id})
                    if edge_id not in collected_edges:
                        edge_in_query_graph = query_graph['edges'][qg_id]
                        source_real_id, target_real_id = id_pair
                        edge_type = upgrade_BiolinkRelation(edge_type)
                        new_edge = {
                            'subject': source_real_id,
                            'object': target_real_id,
                            'predicate': edge_type,
                            'attributes': []
                        }
                        for key, value in edge.items():
                            if key in new_edge:
                                continue
                            new_edge['attributes'].append({
                                'type': 'NA',
                                'name': key,
                                'value': value
                            })
                        collected_edges.add(edge_id)
                        edges_all[edge_id] = new_edge
                        index += 1

                current_answer_bindings['edge_bindings'].update(current_edge_binding)
            answer_bindings += [current_answer_bindings]
        return {"knowledge_graph": {"nodes": nodes_all, "edges": edges_all}, "results": answer_bindings}


if __name__=='__main__':
    q= 'match (a) return count (a); '
    redis_driver = RedisDriver(host='localhost', port='6380', graph_db_name='test')
    import asyncio
    results = asyncio.run(redis_driver.run("""   
    MATCH (n0:`chemical_substance` {`id`: 'CHEBI:39385'})-[e0]-(n1:`named_thing` {}) WITH n0 AS n0, n1 AS n1, collect(e0) AS e0 RETURN n0,n1,e0
    """))
    results


import base64
import traceback

import httpx

from PLATER.services.config import config
from PLATER.services.util.drivers.neo4j_driver import Neo4jHTTPDriver
from PLATER.services.util.drivers.redis_driver import RedisDriver
from PLATER.services.util.logutil import LoggingUtil
from bmt import Toolkit


logger = LoggingUtil.init_logging(__name__,
                                  config.get('logging_level'),
                                  config.get('logging_format')
                                  )

class GraphInterface:
    """
    Singleton class for interfacing with the graph.
    """

    class _GraphInterface:
        def __init__(self, host, port, auth, backend='neo4j', db_name=None):
            if backend == 'neo4j':
                self.driver = Neo4jHTTPDriver(host=host, port=port, auth=auth)
            elif backend == 'redis':
                self.driver = RedisDriver(host=host, port=port, password=auth[1], graph_db_name=db_name)
            self.schema = None
            self.summary = None
            self.toolkit = Toolkit()

        def find_biolink_leaves(self, biolink_concepts: list):
            """
            Given a list of biolink concepts, returns the leaves removing any parent concepts.
            :param biolink_concepts: list of biolink concepts
            :return: leave concepts.
            """
            ancestry_set = set()
            all_concepts = set(biolink_concepts)
            for x in all_concepts:
                ancestors = set(self.toolkit.get_ancestors(x, reflexive=False, formatted=True))
                ancestry_set = ancestry_set.union(ancestors)
            leaf_set = all_concepts - ancestry_set
            return leaf_set

        def get_schema(self, force_update=False):
            """
            Gets the schema of the graph. To be used by. Also generates graph summary
            :return: Dict of structure source label as outer most keys, target labels as inner keys and list of predicates
            as value.
            :rtype: dict
            """
            self.schema_raw_result = {}
            if self.schema is None or force_update:
                query = """
                           MATCH (a)-[x]->(b)                                                                                  
                           RETURN DISTINCT labels(a) as source_labels, type(x) as predicate, labels(b) as target_labels
                           """
                logger.info(f"starting query {query} on graph... this might take a few")
                result = self.driver.run_sync(query)
                logger.info(f"completed query, preparing initial schema")
                structured = self.convert_to_dict(result)
                self.schema_raw_result = structured
                schema_bag = {}
                # permituate source labels and target labels array
                # replacement for unwind for previous cypher
                structured_expanded = []
                for triplet in structured:
                    # Since there are some nodes in data currently just one label ['biolink:NamedThing']
                    # This filter is to avoid that scenario.
                    # @TODO need to remove this filter when data build avoids adding nodes with single ['biolink:NamedThing'] labels.
                    filter_named_thing = lambda x: filter(lambda y: y != 'biolink:NamedThing', x)
                    # For redis convert these to arrays
                    source_labels = [triplet['source_labels']] if isinstance(triplet['source_labels'], str) else triplet['source_labels']
                    target_labels = [triplet['target_labels']] if isinstance(triplet['target_labels'], str) else triplet['target_labels']
                    source_labels, predicate, target_labels = self.find_biolink_leaves(filter_named_thing(source_labels)), \
                                                              triplet['predicate'], \
                                                              self.find_biolink_leaves(filter_named_thing(target_labels))

                    for source_label in source_labels:
                        for target_label in target_labels:
                            structured_expanded.append(
                                {
                                    'source_label': source_label,
                                    'target_label': target_label,
                                    'predicate': predicate
                                }
                            )
                structured = structured_expanded
                for triplet in structured:
                    subject = triplet['source_label']
                    predicate = triplet['predicate']
                    objct = triplet['target_label']
                    if subject not in schema_bag:
                        schema_bag[subject] = {}
                    if objct not in schema_bag[subject]:
                        schema_bag[subject][objct] = []
                    if predicate not in schema_bag[subject][objct]:
                        schema_bag[subject][objct].append(predicate)
                self.schema = schema_bag
                logger.info("schema done.")
                if not self.summary:
                    query = """
                    MATCH (c) RETURN DISTINCT labels(c) as types, count(c) as count                
                    """
                    logger.info(f'generating graph summary: {query}')
                    raw = self.convert_to_dict(self.driver.run_sync(query))
                    summary = {}
                    for node in raw:
                        labels = node['types']
                        labels = labels if isinstance(labels, list) else[labels]
                        count = node['count']
                        query = f"""
                        MATCH (:{':'.join(labels)})-[e]->(b) WITH DISTINCT e , b 
                        RETURN 
                            type(e) as edge_types, 
                            count(e) as edge_counts,
                            labels(b) as target_labels 
                        """
                        raw = self.convert_to_dict(self.driver.run_sync(query))
                        summary_key = ':'.join(labels)
                        summary[summary_key] = {
                            'nodes_count': count
                        }
                        for row in raw:
                            target_label = row['target_labels']
                            target_label = [target_label] if isinstance(target_label, str) else target_label
                            target_key = ':'.join(target_label)
                            edge_name = row['edge_types']
                            edge_count = row['edge_counts']
                            summary[summary_key][target_key] = summary[summary_key].get(target_key, {})
                            summary[summary_key][target_key][edge_name] = edge_count
                    self.summary = summary
                    logger.info(f'generated summary for {len(summary)} node types.')
            return self.schema

        async def get_mini_schema(self, source_id, target_id):
            """
            Given either id of source and/or target returns predicates that relate them. And their
            possible labels.
            :param source_id:
            :param target_id:
            :return:
            """
            source_id_syntaxed = f"{{id: \"{source_id}\"}}" if source_id else ''
            target_id_syntaxed = f"{{id: \"{target_id}\"}}" if target_id else ''
            query = f"""
                            MATCH (a{source_id_syntaxed})-[x]->(b{target_id_syntaxed}) WITH
                                [la in labels(a) where la <> 'Concept'] as source_label,
                                [lb in labels(b) where lb <> 'Concept'] as target_label,
                                type(x) as predicate
                            RETURN DISTINCT source_label, predicate, target_label
                        """
            response = await self.driver.run(query)
            response = self.convert_to_dict(response)
            return response

        async def get_node(self, node_type: str, curie: str) -> list:
            """
            Returns a node that matches curie as its ID.
            :param node_type: Type of the node.
            :type node_type:str
            :param curie: Curie.
            :type curie: str
            :return: value of the node in neo4j.
            :rtype: list
            """
            query = f"MATCH (c:`{node_type}`{{id: '{curie}'}}) return c"
            response = await self.driver.run(query)

            data = response.get('results',[{}])[0].get('data', [])
            '''
            data looks like 
            [
            {'row': [{...node data..}], 'meta': [{...}]},
            {'row': [{...node data..}], 'meta': [{...}]},
            {'row': [{...node data..}], 'meta': [{...}]}
            ]            
            '''
            rows = []
            if len(data):
                from functools import reduce
                rows = reduce(lambda x, y: x + y.get('row', []), data, [])
            return rows

        async def get_single_hops(self, source_type: str, target_type: str, curie: str) -> list:
            """
            Returns a triplets of source to target where source id is curie.
            :param source_type: Type of the source node.
            :type source_type: str
            :param target_type: Type of target node.
            :type target_type: str
            :param curie: Curie of source node.
            :type curie: str
            :return: list of triplets where each item contains source node, edge, target.
            :rtype: list
            """

            query = f'MATCH (c:`{source_type}`{{id: \'{curie}\'}})-[e]->(b:`{target_type}`) return distinct c , e, b'
            response = await self.driver.run(query)
            rows = list(map(lambda data: data['row'], response['results'][0]['data']))
            query = f'MATCH (c:`{source_type}`{{id: \'{curie}\'}})<-[e]-(b:`{target_type}`) return distinct b , e, c'
            response = await self.driver.run(query)
            rows += list(map(lambda data: data['row'], response['results'][0]['data']))

            return rows

        async def run_cypher(self, cypher: str, **kwargs) -> list:
            """
            Runs cypher directly.
            :param cypher: cypher query.
            :type cypher: str
            :return: unprocessed neo4j response.
            :rtype: list
            """
            return await self.driver.run(cypher, **kwargs)

        async def get_sample(self, node_type):
            """
            Returns a few nodes.
            :param node_type: Type of nodes.
            :type node_type: str
            :return: Node dict values.
            :rtype: dict
            """
            query = f"MATCH (c:{node_type}) return c limit 5"
            response = await self.driver.run(query)
            rows = response['results'][0]['data'][0]['row']
            return rows

        async def get_examples(self, source, target=None):
            """
            Returns an example for source node only if target is not specified, if target is specified a sample one hop
            is returned.
            :param source: Node type of the source node.
            :type source: str
            :param target: Node type of the target node.
            :type target: str
            :return: A single source node value if target is not provided. If target is provided too, a triplet.
            :rtype:
            """
            if target:
                query = f"MATCH (source:{source})-[edge]->(target:{target}) return source, edge, target limit 1"
                response = await self.run_cypher(query)
                final = list(map(lambda data: data['row'], response['results'][0]['data']))
                return final
            else:
                query = f"MATCH ({source}:{source}) return {source} limit 1"
                response = await self.run_cypher(query)
                final = list(map(lambda node: node[source], self.driver.convert_to_dict(response)))
                return final

        def supports_apoc(self):
            """
            Returns true if apoc is supported by backend database.
            :return: bool true if neo4j supports apoc.
            """
            return self.driver.check_apoc_support()

        async def run_apoc_cover(self, ids: list):
            """
            Runs apoc.algo.cover on list of ids
            :param ids:
            :return: dictionary of edges and source and target nodes ids
            """
            query = f"""
                        MATCH (node:`biolink:NamedThing`)
                        USING INDEX node:`biolink:NamedThing`(id)
                        WHERE node.id in {ids}
                        WITH collect(node) as nodes
                        CALL apoc.algo.cover(nodes) yield rel
                        WITH {{subject: startNode(rel).id ,
                               object: endNode(rel).id,
                               predicate: type(rel),
                               edge: rel }} as row
                        return collect(row) as result                                        
                        """
            result = self.convert_to_dict(self.driver.run_sync(query))
            return result

        def convert_to_dict(self, result):
            return self.driver.convert_to_dict(result)

        async def answer_trapi_question(self, trapi_question, options={}, timeout=None):
            response = await self.driver.answer_TRAPI_question(trapi_question, options=options, timeout=timeout)
            response.update({'query_graph': trapi_question})
            return response

    instance = None

    def __init__(self, host, port, auth, db_name, db_type):
        # create a new instance if not already created.
        if not GraphInterface.instance:
            GraphInterface.instance = GraphInterface._GraphInterface(host=host, port=port, auth=auth, backend= db_type ,db_name=db_name)

    def __getattr__(self, item):
        # proxy function calls to the inner object.
        return getattr(self.instance, item)



from PLATER.services.util.drivers.redis_trapi_cypher_compiler \
    import cypher_query_answer_map,\
    cypher_prop_string, \
    cypher_query_fragment_match, \
    NodeReference, \
    EdgeReference


def test_node_reference():
    node = {
        'category': 'chemical_substance',
        'id': 'CURIE:1',
        'is_set': False
    }
    cypher_node = NodeReference(node, node_id='xxxx')

    node_rep = cypher_node.__str__()
    assert node_rep == f'{cypher_node.name}' + f'{cypher_node.prop_string}'
    node_rep = cypher_node.__str__()
    assert  node_rep == cypher_node.name
    assert "'chemical_substance' in xxxx.category" in cypher_node.filters
    assert "'biolink:ChemicalSubstance' in xxxx.category" in cypher_node.filters
    assert cypher_node.prop_string == " {`id`: 'CURIE:1'}"
    assert cypher_node.name == 'xxxx'
    assert cypher_node.labels == [node['category']]
    # node with multiple extra vars and types
    node = {
        'category': ['chemical_substance', 'named_thing'],
        'id': ['CURIE:1', 'CURIE:2'],
        'extra_prop': 'prop1',
        'is_set': False
    }
    cypher_node = NodeReference(node, node_id='n0')
    node_rep = cypher_node.__str__()
    assert node_rep == f'{cypher_node.name}' + f'{cypher_node.prop_string}'
    node_rep = cypher_node.__str__()
    assert node_rep == cypher_node.name
    split_by_and = cypher_node.filters.split(' AND ')
    # note for performance we should go for curie filter first then category (since looking up array is expensive?)
    all_filters = list(map(lambda x: x.strip(' '), split_by_and[0].strip('(').strip(')').split(' OR ')))
    assert "n0.id = 'CURIE:1'" in all_filters
    assert "n0.id = 'CURIE:2'" in all_filters
    all_filters = list(map(lambda x: x.strip(' '), split_by_and[1].strip('(').strip(')').split(' OR ')))
    assert "'chemical_substance' in n0.category" in all_filters
    assert "'named_thing' in n0.category" in all_filters
    assert "'biolink:ChemicalSubstance' in n0.category" in all_filters
    assert "'biolink:NamedThing' in n0.category" in all_filters
    assert cypher_node.prop_string == ' {`extra_prop`: \'prop1\'}'

def test_edge_reference():
    # typed edge
    edge_id = 'e0'
    edge = {
        'predicate': 'biolink:some_edge',
        'subject': 'n0',
        'object': 'n1'
    }
    cypher_edge = EdgeReference(edge, edge_id=edge_id)
    # first access returns match pattern
    assert f'-[{edge_id}:`{edge["predicate"]}`]->' == str(cypher_edge)
    # second access returns with out type
    assert f'-[{edge_id}]->' == str(cypher_edge)
    assert cypher_edge.filters == ''
    # when edges have multiple types
    edge_id = 'e0'
    edge = {
        'predicate': ['biolink:some_edge', 'biolink:other_type', 'third_type'],
        'subject': 'n0',
        'object': 'n1'
    }
    cypher_edge = EdgeReference(edge,edge_id=edge_id)
    # first and second access should stay same
    assert f'-[{edge_id}]->' == str(cypher_edge)
    assert f'-[{edge_id}]->' == str(cypher_edge)
    all_filters = list(map(lambda x: x.strip(' '), cypher_edge.filters.split(' OR ')))
    assert f'type(e0) = "biolink:some_edge"' in all_filters
    assert f'type(e0) = "some_edge"' in all_filters
    assert f'type(e0) = "biolink:other_type"' in all_filters
    assert f'type(e0) = "other_type"' in all_filters
    assert f'type(e0) = "biolink:third_type"' in all_filters
    assert f'type(e0) = "third_type"' in all_filters



def test_match_header():
    trapi_question = {
        'nodes': {
            'n0': {'category': 'x', 'id': 'y'},
            'n1': {'category': 'y'},
        }, 'edges': {
            'e0':{'subject': 'n0', 'object': 'n1', 'predicate': ['biolink:type', 'biolink:Other']}
        }
    }
    match_clause = cypher_query_fragment_match(trapi_question)
    cypher_nodes = list(map(lambda x: NodeReference(trapi_question['nodes'][x], node_id=x), trapi_question['nodes']))
    cypher_edges = list(map(lambda x: EdgeReference(trapi_question['edges'][x], edge_id=x), trapi_question['edges']))
    match_head = match_clause.split('\n')[0]
    for node in cypher_nodes:
        assert node.__str__() in match_head
    for edge in cypher_edges:
        assert edge.__str__() in match_head


def test_where_conditions():
    trapi_question = {
        'nodes': {
            'n0': {'category': 'x', 'id': 'y'},
            'n1': {'category': 'y'},
        }, 'edges': {
            'e0':{'subject': 'n0', 'object': 'n1', 'predicate': ['biolink:type', 'biolink:Other']}
        }
    }
    match_fragment = cypher_query_fragment_match(trapi_question)
    where_clauses = match_fragment.split('WHERE')[1]
    cypher_nodes = list(map(lambda x: NodeReference(trapi_question['nodes'][x], x), trapi_question['nodes']))
    cypher_edges = list(map(lambda x: EdgeReference(trapi_question['edges'][x], x), trapi_question['edges']))
    all_filters = []
    for node in cypher_nodes:
        assert node.filters in where_clauses
        all_filters.append(f'({node.filters})')
    for edge in cypher_edges:
        assert edge.filters in where_clauses
        all_filters.append(f'({edge.filters})')
    # make sure all edge and node filters are joined with AND
    where_split_with_and = list(map(lambda x: x.strip(' '), where_clauses.split('\nAND ')))
    assert len(all_filters) == len(where_split_with_and)
    for f in all_filters:
        assert f in where_split_with_and


def test_returns():
    trapi_question = {
        'nodes': {
            'n0':{'category': 'x', 'id': 'y'},
            'n1':{'category': 'y', 'is_set': True}
        }, 'edges': {
             'e0':{'subject': 'n0', 'object': 'n1', 'predicate': ['biolink:type', 'biolink:Other']}
        }
    }
    match_fragment = cypher_query_fragment_match(trapi_question)
    whole_cypher = cypher_query_answer_map(trapi_question)
    returns = whole_cypher.replace(match_fragment +'\n', '').split('\n')
    with_clause, returns_clause = list(map(lambda x: x.strip(' '), returns[0].replace('WITH ', '').split(',')))\
        , list(map(lambda x: x.strip(' '), returns[1].replace('RETURN','').split(',')))
    # make sure normal nodes are return
    assert 'n0 AS n0' in with_clause
    # make sure sets are collected
    assert 'collect(n1) AS n1' in with_clause
    # make sure edges are collected aswell
    assert 'collect(e0) AS e0' in with_clause

    assert 'n0' in returns_clause
    assert 'n1' in returns_clause
    assert 'e0' in returns_clause

    # check for types
    assert 'labels(n0) AS type__n0' in returns_clause
    # types for collected node should be an array
    assert '[node in n1 | labels(node)] AS type__n1' in returns_clause
    # types for edges is similar
    assert '[edge in e0 | type(edge)] AS type__e0' in returns_clause
    first_part , second_part = '[edge in e0 | [startNode(edge).id, endNode(edge).id]] AS id_pairs__e0'.split(',')
    assert first_part.strip() in returns_clause
    assert second_part.strip() in returns_clause




# This test will test a running instance of plater with
# running cypher on the cypher endpoint and asserting results
# to equivalent TRAPI messages
#

import pytest
import os
from functools import reduce
import requests

PLATERURL = os.environ.get('PLATER_HOST', 'http://localhost:81')
if not PLATERURL:
    raise Exception("Error : Please set PLATER_HOST env var")

@pytest.fixture
def connected_nodes_ids():
    connected_node_cypher = {
        "query": "MATCH (a)--() return distinct a limit 20"
    }

    response = requests.post(PLATERURL + '/cypher', json=connected_node_cypher)
    assert response.status_code == 200, f"ERROR running cypher to endpoint {PLATERURL}/cypher"
    response = response.json()
    results = response['results'][0]
    cols = results['columns']
    rows = reduce(lambda x, y: x + y, [d['row'] for d in results['data']], [])
    ids = list(map(lambda x : x['id'], rows))
    return ids


def convert_to_dict(response: dict) -> list:
    """
    Converts a neo4j result to a structured result.
    :param response: neo4j http raw result.
    :type response: dict
    :return: reformatted dict
    :rtype: dict
    """
    results = response.get('results')
    array = []
    if results:
        for result in results:
            cols = result.get('columns')
            if cols:
                data_items = result.get('data')
                for item in data_items:
                    new_row = {}
                    row = item.get('row')
                    for col_name, col_value in zip(cols, row):
                        new_row[col_name] = col_value
                    array.append(new_row)
    return array

def test_single_hop(connected_nodes_ids):
    for i in connected_nodes_ids:
        cypher_query = {
            "query": f"MATCH (a{{id : '{i}'}})-[e]-(b) return a, e, b"
        }
        cypher_resp = requests.post(PLATERURL + '/cypher', json=cypher_query)
        assert cypher_resp.status_code == 200
        cypher_resp = cypher_resp.json()
        cypher_resp = convert_to_dict(cypher_resp)
        query_graph = {
            "query_graph":{
                'nodes': [
                    {'id': 'a', 'curie': f"{i}", 'type': 'named_thing'},
                    {'id': 'b', 'type': 'named_thing'}
                ],
                'edges': [
                    {'id': 'e', 'source_id': 'a', 'target_id': 'b'}
                ]
            }
        }
        import json
        print(json.dumps({"message":query_graph}, indent=2))
        trapi_response = requests.post(PLATERURL + '/query', json={'message': query_graph})
        assert trapi_response.status_code == 200
        trapi_response = trapi_response.json()
        assert len(trapi_response['knowledge_graph']['nodes']) != 0
        assert len(trapi_response['knowledge_graph']['edges']) != 0
        assert len(trapi_response['results'])
        path_array = []
        # check edges and nodes exist in knowledge graph
        for row in cypher_resp:
            found_a = False
            found_b = False
            found_e = False
            for node in trapi_response['knowledge_graph']['nodes']:
                if row['a']['id'] == node['id']:
                    found_a = True
                    for key in row['a'].keys():
                        assert row['a'][key] == node[key]
            assert found_a
            for node in trapi_response['knowledge_graph']['nodes']:
                if row['b']['id'] == node['id']:
                    found_b = True
                    for key in row['b'].keys():
                        assert row['b'][key] == node[key]
            assert found_b
            for edge in trapi_response['knowledge_graph']['edges']:
                if row['e']['id'] == edge['id']:
                    found_e = True
                    for key in row['e'].keys():
                        assert row['e'][key] == edge[key]
            assert found_e
            path_array.append([row['a']['id'], row['e']['id'], row['b']['id'] ])
        for result in trapi_response['results']:
            assert result['node_bindings']
            assert result['edge_bindings']
            binding_as_dict = {r['qg_id']: r['kg_id'] for r in result['node_bindings']}
            binding_as_dict.update({r['qg_id']: r['kg_id'] for r in result['edge_bindings']})
            current_path = [binding_as_dict['a'], binding_as_dict['e'], binding_as_dict['b']]
            assert current_path in path_array



import os
import re
import sys
from shutil import rmtree
from setuptools import setup, find_packages, Command

current = os.path.abspath(os.path.dirname(__file__))


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

version = "1.9.8"

class PublishClass(Command):
    description = "Publish the package"
    user_options = []

    # This method must be implemented
    def initialize_options(self):
        pass

    # This method must be implemented
    def finalize_options(self):
        pass

    def run(self):
        try:
            print(f"-----> removing previous builds")
            rmtree(os.path.join(current, 'dist'))
            rmtree(os.path.join(current, 'build'))
        except Exception as e:
            print(f"-----> Exception: {e}")
            pass
        os.system('python setup.py sdist bdist_wheel --universal')
        # os.system('twine upload --repository dist/*')
        sys.exit()

setup(
    name="PLATER-GRAPH",
    version=version,
    maintainer="Renci",
    maintainer_email="yaphetkg@renci.org",
    description="Graph DB interface for Translator API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yaphetkg/plater.git",
    # package_dir={"": 'PLATER'},
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "coverage",
        "pyaml==20.3.1",
        "pytest==5.4.1",
        "pytest-asyncio==0.14.0",
        "starlette==0.13.6",
        "uvicorn==0.11.7",
        "httpx",
        "redis"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    dependency_links= [
        "git+git://github.com/patrickkwang/biolink-model-toolkit@master#egg=bmt",
        "git+https://github.com/ranking-agent/reasoner.git",
        "git+https://github.com/TranslatorSRI/reasoner-pydantic@v1.0#egg=reasoner-pydantic",
        "git+https://github.com/patrickkwang/fastapi#egg=fastapi",
        "git+https://github.com/redislabs/redisgraph-py.git"
    ],
    python_requires='>=3.7',
    cmdclass={
        'publish': PublishClass,
    },
)


import asyncio
import pytest
from PLATER.services.util.overlay import Overlay


@pytest.fixture()
def graph_interface_apoc_supported():
    class MockGI:
        def supports_apoc(self):
            return True

        async def run_apoc_cover(self, idlist):
            return [{
                'result': [
                    {
                        'subject': 'NODE:0',
                        'object': 'NODE:2',
                        'predicate': 'biolink:related_to',
                        'edge': {
                            'id': 'SUPPORT_EDGE_KG_ID_1',
                            'attr_1': [],
                            'attr_2': {}
                        }
                    }, {
                        'subject': 'NODE:00',
                        'object': 'NODE:22',
                        'predicate': 'biolink:related_to',
                        'edge': {
                            'type': 'biolink:related_to',
                            'id': 'SUPPORT_EDGE_KG_ID_2',
                            'attr_1': [],
                            'attr_2': {}
                        }
                    }, {  # Edge relating two nodes from different answers
                        # we should expect this NOT to be in response.
                        'subject': 'NODE:0',
                        'object': 'NODE:22',
                        'predicate': 'biolink:related_to',
                        'edge': {
                            'type': 'biolink:related_to',
                            'id': 'SUPPORT_EDGE_KG_ID_3'
                        }
                    }
                ]
            }]

    return MockGI()


@pytest.fixture()
def graph_interface_apoc_unsupported():
    class MockGI:
        def supports_apoc(self):
            return True

    return MockGI()


@pytest.fixture()
def reasoner_json():
    return {
        # Although this is not particularly useful in testing...
        'query_graph': {
            'nodes': {
                'n0': {'type': 'type'},
                'n1': {'type': 'type'},
                'n2': {'type': 'type'}
            },
            'edges':{
                'e0': {'subject': 'n0', 'object': 'n1'},
                'e1': {'subject': 'n1', 'object': 'n2'},
            }
        },
        # Knowledge_graph Here also we don't really care about what was in
        # kg
        'knowledge_graph':
            {
                'nodes': {},
                'edges': {}
            },
        'results': [
            {
                'node_bindings': {
                    'n0': [{'id': 'NODE:0'}],
                    'n1': [{'id': 'NODE:1'}],
                    'n2': [{'id': 'NODE:2'}],
                },
                'edge_bindings': {
                    'e0': [{'id': 'EDGE:0'}],
                    'e1': [{'id': 'EDGE:1'}],
                    'e2': [{'id': 'EDGE:2'}]
                }
            },
            {
                'node_bindings': {
                    'n0': [{'id': 'NODE:00'}],
                    'n1': [{'id': 'NODE:11'}],
                    'n2': [{'id': 'NODE:22'}],
                },
                'edge_bindings': {
                    'e0': [{'id': 'EDGE:00'}],
                    'e1': [{'id': 'EDGE:11'}],
                    'e2': [{'id': 'EDGE:22'}]
                }
            },
            {
                'node_bindings': {
                    'n0': [{'id': 'NODE:000'}],
                    'n1': [{'id': 'NODE:111'}],
                    'n2': [{'id': 'NODE:222'}],
                },
                'edge_bindings': {
                    'e0': [{'id': 'EDGE:000'}],
                    'e1': [{'id': 'EDGE:111'}],
                    'e2': [{'id': 'EDGE:222'}]
                },
            }
        ]
    }


def get_kg_ids(bindings):
    all_ids = []
    for key in bindings:
        items = bindings[key]
        for item in items:
            all_ids.append(item['id'])
    return all_ids


def test_overlay_adds_support_bindings(graph_interface_apoc_supported, reasoner_json):
    ov = Overlay(graph_interface=graph_interface_apoc_supported)
    event_loop = asyncio.get_event_loop()
    response = event_loop.run_until_complete(ov.overlay_support_edges(reasoner_json))
    edges = response['knowledge_graph']['edges']
    edge_ids = edges.keys()
    assert len(edge_ids) == 2
    assert 'SUPPORT_EDGE_KG_ID_1' in edge_ids
    assert 'SUPPORT_EDGE_KG_ID_2' in edge_ids
    checked = False
    for answer in response['results']:
        node_bindings = answer['node_bindings']
        all_node_ids = get_kg_ids(node_bindings)
        edge_bindings = answer['edge_bindings']
        all_edge_kg_ids = get_kg_ids(edge_bindings)
        if ('NODE:0' in all_node_ids and 'NODE:2' in all_node_ids) \
                or ('NODE:00' in all_node_ids and 'NODE:22' in all_node_ids):
            assert 'SUPPORT_EDGE_KG_ID_1' in all_edge_kg_ids or 'SUPPORT_EDGE_KG_ID_2' in all_edge_kg_ids
            checked = True
    assert checked
